"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var pluralize = require("pluralize");
var Type_1 = require("./Type");
var TypeNames_1 = require("./TypeNames");
function gatherNames(graph) {
    function setNames(t, tn) {
        graph.attributeStore.set(TypeNames_1.namesTypeAttributeKind, t, tn);
    }
    graph.allTypesUnordered().forEach(function (t) {
        if (t.hasNames) {
            setNames(t, t.getNames().clearInferred());
        }
    });
    var processed = immutable_1.Set();
    function processType(t, names, parentNames, alternativeSuffix) {
        if (t.hasNames) {
            var typeNames = t.getNames();
            if (!typeNames.areInferred) {
                names = typeNames.names;
            }
        }
        var alternatives = [];
        names.forEach(function (name) {
            if (alternativeSuffix !== undefined) {
                // FIXME: we should only add these for names we couldn't singularize
                alternatives.push(name + "_" + alternativeSuffix);
            }
            if (parentNames !== undefined) {
                alternatives.push.apply(alternatives, parentNames.map(function (pn) { return pn + "_" + name; }).toArray());
                // FIXME: add alternatives with the suffix here, too?
            }
            alternatives.push(name + "_" + t.kind);
            if (parentNames !== undefined) {
                alternatives.push.apply(alternatives, parentNames.map(function (pn) { return pn + "_" + name + "_" + t.kind; }).toArray());
                // FIXME: add alternatives with the suffix here, too?
            }
        });
        var newNames = new TypeNames_1.TypeNames(names, immutable_1.OrderedSet(alternatives), true);
        setNames(t, t.hasNames ? t.getNames().add(newNames) : newNames);
        var processedEntry = immutable_1.List([t, names, parentNames]);
        if (processed.has(processedEntry))
            return;
        processed = processed.add(processedEntry);
        Type_1.matchCompoundType(t, function (arrayType) {
            processType(arrayType.items, names.map(pluralize.singular), parentNames, "element");
        }, function (classType) {
            var properties = classType.properties.sortBy(function (_, n) { return n; });
            properties.forEach(function (property, propertyName) {
                processType(property.type, immutable_1.OrderedSet([propertyName]), names);
            });
        }, function (mapType) {
            processType(mapType.values, names.map(pluralize.singular), parentNames, "value");
        }, function (unionType) {
            var members = unionType.members.sortBy(function (member) { return member.kind; });
            members.forEach(function (memberType) {
                processType(memberType, names, parentNames);
            });
        });
    }
    graph.topLevels.forEach(function (t, name) {
        processType(t, immutable_1.OrderedSet([name]), undefined);
    });
}
exports.gatherNames = gatherNames;
