"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var TypeGraph_1 = require("./TypeGraph");
var TypeAttributes_1 = require("./TypeAttributes");
var Support_1 = require("./Support");
var TypeRef = /** @class */ (function () {
    function TypeRef(graph, index, _allocatingTypeBuilder) {
        this.graph = graph;
        this._allocatingTypeBuilder = _allocatingTypeBuilder;
        this._maybeIndexOrRef = index;
    }
    TypeRef.prototype.follow = function () {
        if (this._maybeIndexOrRef instanceof TypeRef) {
            return this._maybeIndexOrRef.follow();
        }
        return this;
    };
    Object.defineProperty(TypeRef.prototype, "maybeIndex", {
        get: function () {
            var tref = this.follow();
            if (typeof tref._maybeIndexOrRef === "number") {
                return tref._maybeIndexOrRef;
            }
            return undefined;
        },
        enumerable: true,
        configurable: true
    });
    TypeRef.prototype.getIndex = function () {
        var maybeIndex = this.maybeIndex;
        if (maybeIndex === undefined) {
            var tref = this.follow();
            if (tref._allocatingTypeBuilder !== undefined) {
                var allocated = tref._allocatingTypeBuilder.reserveTypeRef();
                Support_1.assert(allocated.follow() !== this, "Tried to create a TypeRef cycle");
                tref._maybeIndexOrRef = allocated;
                tref._allocatingTypeBuilder = undefined;
                return allocated.getIndex();
            }
            return Support_1.panic("Trying to dereference unresolved type reference");
        }
        return maybeIndex;
    };
    TypeRef.prototype.callWhenResolved = function (callback) {
        if (this._maybeIndexOrRef === undefined) {
            if (this._callbacks === undefined) {
                this._callbacks = [];
            }
            this._callbacks.push(callback);
        }
        else if (typeof this._maybeIndexOrRef === "number") {
            callback(this._maybeIndexOrRef);
        }
        else {
            this._maybeIndexOrRef.callWhenResolved(callback);
        }
    };
    TypeRef.prototype.resolve = function (tref) {
        if (this._maybeIndexOrRef !== undefined) {
            Support_1.assert(this.maybeIndex === tref.maybeIndex, "Trying to resolve an allocated type reference with an incompatible one");
        }
        Support_1.assert(tref.follow() !== this, "Tried to create a TypeRef cycle");
        this._maybeIndexOrRef = tref.follow();
        this._allocatingTypeBuilder = undefined;
        if (this._callbacks !== undefined) {
            for (var _i = 0, _a = this._callbacks; _i < _a.length; _i++) {
                var cb = _a[_i];
                tref.callWhenResolved(cb);
            }
            this._callbacks = undefined;
        }
    };
    TypeRef.prototype.deref = function () {
        return this.graph.atIndex(this.getIndex());
    };
    TypeRef.prototype.equals = function (other) {
        if (!(other instanceof TypeRef)) {
            return false;
        }
        Support_1.assert(this.graph === other.graph, "Comparing type refs of different graphs");
        return this.follow() === other.follow();
    };
    TypeRef.prototype.hashCode = function () {
        return this.getIndex() | 0;
    };
    return TypeRef;
}());
exports.TypeRef = TypeRef;
function provenanceToString(p) {
    return p
        .map(function (r) { return r.getIndex(); })
        .toList()
        .sort()
        .map(function (i) { return i.toString(); })
        .join(",");
}
exports.provenanceTypeAttributeKind = new TypeAttributes_1.TypeAttributeKind("provenance", Support_1.setUnion, function (a) { return a; }, provenanceToString);
exports.NoStringTypeMapping = {
    date: "date",
    time: "time",
    dateTime: "date-time"
};
var TypeBuilder = /** @class */ (function () {
    function TypeBuilder(_stringTypeMapping, alphabetizeProperties, _allPropertiesOptional, _addProvenanceAttributes, inheritsProvenanceAttributes) {
        var _this = this;
        this._stringTypeMapping = _stringTypeMapping;
        this.alphabetizeProperties = alphabetizeProperties;
        this._allPropertiesOptional = _allPropertiesOptional;
        this._addProvenanceAttributes = _addProvenanceAttributes;
        this.topLevels = immutable_1.Map();
        this.types = [];
        this.typeAttributes = [];
        this.commitType = function (tref, t) {
            var index = tref.getIndex();
            // const name = names !== undefined ? ` ${names.combinedName}` : "";
            // console.log(`committing ${t.kind}${name} to ${index}`);
            Support_1.assert(_this.types[index] === undefined, "A type index was committed twice");
            _this.types[index] = t;
        };
        // FIXME: make mutable?
        this._primitiveTypes = immutable_1.Map();
        this._noEnumStringType = undefined;
        this._mapTypes = immutable_1.Map();
        this._arrayTypes = immutable_1.Map();
        this._enumTypes = immutable_1.Map();
        this._classTypes = immutable_1.Map();
        this._unionTypes = immutable_1.Map();
        Support_1.assert(!_addProvenanceAttributes || !inheritsProvenanceAttributes, "We can't both inherit as well as add provenance");
        this.typeGraph = new TypeGraph_1.TypeGraph(this, _addProvenanceAttributes || inheritsProvenanceAttributes);
    }
    TypeBuilder.prototype.addTopLevel = function (name, tref) {
        // assert(t.typeGraph === this.typeGraph, "Adding top-level to wrong type graph");
        Support_1.assert(!this.topLevels.has(name), "Trying to add top-level with existing name");
        Support_1.assert(this.types[tref.getIndex()] !== undefined, "Trying to add a top-level type that doesn't exist (yet?)");
        this.topLevels = this.topLevels.set(name, tref);
    };
    TypeBuilder.prototype.reserveTypeRef = function () {
        var index = this.types.length;
        // console.log(`reserving ${index}`);
        this.types.push(undefined);
        var tref = new TypeRef(this.typeGraph, index, undefined);
        var attributes = this._addProvenanceAttributes
            ? exports.provenanceTypeAttributeKind.makeAttributes(immutable_1.Set([tref]))
            : immutable_1.Map();
        this.typeAttributes.push(attributes);
        return tref;
    };
    TypeBuilder.prototype.addType = function (forwardingRef, creator, attributes) {
        if (forwardingRef !== undefined && forwardingRef.maybeIndex !== undefined) {
            Support_1.assert(this.types[forwardingRef.maybeIndex] === undefined);
        }
        var tref = forwardingRef !== undefined && forwardingRef.maybeIndex !== undefined
            ? forwardingRef
            : this.reserveTypeRef();
        if (attributes !== undefined) {
            this.addAttributes(tref, attributes);
        }
        var t = creator(tref);
        this.commitType(tref, t);
        if (forwardingRef !== undefined && tref !== forwardingRef) {
            forwardingRef.resolve(tref);
        }
        return tref;
    };
    TypeBuilder.prototype.atIndex = function (index) {
        var maybeType = this.types[index];
        if (maybeType === undefined) {
            return Support_1.panic("Trying to deref an undefined type in a type builder");
        }
        var maybeNames = this.typeAttributes[index];
        return [maybeType, maybeNames];
    };
    TypeBuilder.prototype.addAttributes = function (tref, attributes) {
        var _this = this;
        tref.callWhenResolved(function (index) {
            if (attributes === undefined) {
                attributes = immutable_1.Map();
            }
            _this.typeAttributes[index] = TypeAttributes_1.combineTypeAttributes(_this.typeAttributes[index], attributes);
        });
    };
    TypeBuilder.prototype.makeNullable = function (tref, attributes) {
        var t = Support_1.defined(this.types[tref.getIndex()]);
        if (t.kind === "null" || t.kind === "any") {
            return tref;
        }
        var nullType = this.getPrimitiveType("null");
        if (!(t instanceof Type_1.UnionType)) {
            return this.getUnionType(attributes, immutable_1.OrderedSet([tref, nullType]));
        }
        var _a = Type_1.removeNullFromUnion(t), maybeNull = _a[0], nonNulls = _a[1];
        if (maybeNull !== null)
            return tref;
        return this.getUnionType(attributes, nonNulls.map(function (nn) { return nn.typeRef; }).add(nullType));
    };
    TypeBuilder.prototype.finish = function () {
        this.typeGraph.freeze(this.topLevels, this.types.map(Support_1.defined), this.typeAttributes);
        return this.typeGraph;
    };
    TypeBuilder.prototype.getPrimitiveType = function (kind, forwardingRef) {
        Support_1.assert(kind !== "string", "Use getStringType to create strings");
        if (kind === "date")
            kind = this._stringTypeMapping.date;
        if (kind === "time")
            kind = this._stringTypeMapping.time;
        if (kind === "date-time")
            kind = this._stringTypeMapping.dateTime;
        if (kind === "string") {
            return this.getStringType(undefined, undefined, forwardingRef);
        }
        var tref = this._primitiveTypes.get(kind);
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.PrimitiveType(tr, kind); }, undefined);
            this._primitiveTypes = this._primitiveTypes.set(kind, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.getStringType = function (attributes, cases, forwardingRef) {
        if (cases === undefined) {
            // FIXME: Right now we completely ignore names for strings
            // without enum cases.  That's the correct behavior at the time,
            // because string types never are assigned names, but we might
            // do that at some point, but in that case we'll want a different
            // type for each occurrence, not the same single string type with
            // all the names.
            //
            // The proper solution at that point might be to just figure
            // out whether we do want string types to have names (we most
            // likely don't), and if not, still don't keep track of them.
            if (this._noEnumStringType === undefined) {
                this._noEnumStringType = this.addType(forwardingRef, function (tr) { return new Type_1.StringType(tr, undefined); }, undefined);
            }
            this.addAttributes(this._noEnumStringType, attributes);
            return this._noEnumStringType;
        }
        return this.addType(forwardingRef, function (tr) { return new Type_1.StringType(tr, cases); }, attributes);
    };
    TypeBuilder.prototype.getEnumType = function (attributes, cases, forwardingRef) {
        var unorderedCases = cases.toSet();
        var tref = this._enumTypes.get(unorderedCases);
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.EnumType(tr, cases); }, attributes);
            this._enumTypes = this._enumTypes.set(unorderedCases, tref);
        }
        else {
            this.addAttributes(tref, attributes);
        }
        return tref;
    };
    TypeBuilder.prototype.getMapType = function (values, forwardingRef) {
        var tref = this._mapTypes.get(values);
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.MapType(tr, values); }, undefined);
            this._mapTypes = this._mapTypes.set(values, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.getArrayType = function (items, forwardingRef) {
        var tref = this._arrayTypes.get(items);
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.ArrayType(tr, items); }, undefined);
            this._arrayTypes = this._arrayTypes.set(items, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.modifyPropertiesIfNecessary = function (properties) {
        if (this.alphabetizeProperties) {
            properties = properties.sortBy(function (_, n) { return n; });
        }
        if (this._allPropertiesOptional) {
            properties = properties.map(function (cp) { return new Type_1.ClassProperty(cp.typeRef, true); });
        }
        return properties;
    };
    TypeBuilder.prototype.getClassType = function (attributes, properties, forwardingRef) {
        properties = this.modifyPropertiesIfNecessary(properties);
        var tref = this._classTypes.get(properties.toMap());
        // FIXME: It's not clear to me that the `forwardingRef` condition here
        // might actually ever be true.  And if it can, shouldn't we also have
        // it in all the other `getXXX` methods here?
        if ((forwardingRef !== undefined && forwardingRef.maybeIndex !== undefined) || tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.ClassType(tr, false, properties); }, attributes);
            this._classTypes = this._classTypes.set(properties.toMap(), tref);
        }
        else {
            this.addAttributes(tref, attributes);
        }
        return tref;
    };
    // FIXME: Maybe just distinguish between this and `getClassType`
    // via a flag?  That would make `ClassType.map` simpler.
    TypeBuilder.prototype.getUniqueClassType = function (attributes, isFixed, properties, forwardingRef) {
        if (properties !== undefined) {
            properties = this.modifyPropertiesIfNecessary(properties);
        }
        return this.addType(forwardingRef, function (tref) { return new Type_1.ClassType(tref, isFixed, properties); }, attributes);
    };
    TypeBuilder.prototype.setClassProperties = function (ref, properties) {
        var type = ref.deref()[0];
        if (!(type instanceof Type_1.ClassType)) {
            return Support_1.panic("Tried to set properties of non-class type");
        }
        properties = this.modifyPropertiesIfNecessary(properties);
        type.setProperties(properties);
    };
    TypeBuilder.prototype.getUnionType = function (attributes, members, forwardingRef) {
        var unorderedMembers = members.toSet();
        var tref = this._unionTypes.get(unorderedMembers);
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.UnionType(tr, members); }, attributes);
            this._unionTypes = this._unionTypes.set(unorderedMembers, tref);
        }
        else {
            this.addAttributes(tref, attributes);
        }
        return tref;
    };
    TypeBuilder.prototype.getUniqueUnionType = function (attributes, members, forwardingRef) {
        return this.addType(forwardingRef, function (tref) { return new Type_1.UnionType(tref, members); }, attributes);
    };
    TypeBuilder.prototype.getUniqueIntersectionType = function (attributes, members, forwardingRef) {
        return this.addType(forwardingRef, function (tref) { return new Type_1.IntersectionType(tref, members); }, attributes);
    };
    TypeBuilder.prototype.setSetOperationMembers = function (ref, members) {
        var type = ref.deref()[0];
        if (!(type instanceof Type_1.UnionType || type instanceof Type_1.IntersectionType)) {
            return Support_1.panic("Tried to set members of non-set-operation type");
        }
        type.setMembers(members);
    };
    return TypeBuilder;
}());
exports.TypeBuilder = TypeBuilder;
var TypeGraphBuilder = /** @class */ (function (_super) {
    __extends(TypeGraphBuilder, _super);
    function TypeGraphBuilder() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    TypeGraphBuilder.prototype.typeForEntry = function (entry) {
        return entry;
    };
    TypeGraphBuilder.prototype.getLazyMapType = function (valuesCreator) {
        return this.addType(undefined, function (tref) { return new Type_1.MapType(tref, valuesCreator()); }, undefined);
    };
    TypeGraphBuilder.prototype.setLostTypeAttributes = function () {
        return;
    };
    return TypeGraphBuilder;
}(TypeBuilder));
exports.TypeGraphBuilder = TypeGraphBuilder;
var TypeReconstituter = /** @class */ (function () {
    function TypeReconstituter(_typeBuilder, _makeClassUnique, _typeAttributes, _forwardingRef) {
        this._typeBuilder = _typeBuilder;
        this._makeClassUnique = _makeClassUnique;
        this._typeAttributes = _typeAttributes;
        this._forwardingRef = _forwardingRef;
        this._wasUsed = false;
    }
    TypeReconstituter.prototype.useBuilder = function () {
        Support_1.assert(!this._wasUsed, "TypeReconstituter used more than once");
        this._wasUsed = true;
        return this._typeBuilder;
    };
    TypeReconstituter.prototype.addAttributes = function (tref) {
        this._typeBuilder.addAttributes(tref, this._typeAttributes);
        return tref;
    };
    TypeReconstituter.prototype.getPrimitiveType = function (kind) {
        return this.addAttributes(this.useBuilder().getPrimitiveType(kind, this._forwardingRef));
    };
    TypeReconstituter.prototype.getStringType = function (enumCases) {
        return this.useBuilder().getStringType(this._typeAttributes, enumCases, this._forwardingRef);
    };
    TypeReconstituter.prototype.getEnumType = function (cases) {
        return this.useBuilder().getEnumType(Support_1.defined(this._typeAttributes), cases, this._forwardingRef);
    };
    TypeReconstituter.prototype.getMapType = function (values) {
        return this.addAttributes(this.useBuilder().getMapType(values, this._forwardingRef));
    };
    TypeReconstituter.prototype.getArrayType = function (items) {
        return this.addAttributes(this.useBuilder().getArrayType(items, this._forwardingRef));
    };
    TypeReconstituter.prototype.getClassType = function (properties) {
        if (this._makeClassUnique) {
            return this.getUniqueClassType(false, properties);
        }
        return this.useBuilder().getClassType(Support_1.defined(this._typeAttributes), properties, this._forwardingRef);
    };
    TypeReconstituter.prototype.getUniqueClassType = function (isFixed, properties) {
        return this.useBuilder().getUniqueClassType(Support_1.defined(this._typeAttributes), isFixed, properties, this._forwardingRef);
    };
    TypeReconstituter.prototype.getUnionType = function (members) {
        return this.useBuilder().getUnionType(Support_1.defined(this._typeAttributes), members, this._forwardingRef);
    };
    TypeReconstituter.prototype.getUniqueIntersectionType = function (members) {
        return this.useBuilder().getUniqueIntersectionType(Support_1.defined(this._typeAttributes), members, this._forwardingRef);
    };
    return TypeReconstituter;
}());
exports.TypeReconstituter = TypeReconstituter;
var GraphRewriteBuilder = /** @class */ (function (_super) {
    __extends(GraphRewriteBuilder, _super);
    function GraphRewriteBuilder(_originalGraph, stringTypeMapping, alphabetizeProperties, graphHasProvenanceAttributes, setsToReplace, _replacer) {
        var _this = _super.call(this, stringTypeMapping, alphabetizeProperties, false, false, graphHasProvenanceAttributes) || this;
        _this._originalGraph = _originalGraph;
        _this._replacer = _replacer;
        _this._reconstitutedTypes = immutable_1.Map();
        _this._reconstitutedUnions = immutable_1.Map();
        _this._lostTypeAttributes = false;
        _this._setsToReplaceByMember = immutable_1.Map();
        var _loop_1 = function (types) {
            var set = immutable_1.Set(types);
            set.forEach(function (t) {
                var index = t.typeRef.getIndex();
                Support_1.assert(!_this._setsToReplaceByMember.has(index), "A type is member of more than one set to be replaced");
                _this._setsToReplaceByMember = _this._setsToReplaceByMember.set(index, set);
            });
        };
        for (var _i = 0, setsToReplace_1 = setsToReplace; _i < setsToReplace_1.length; _i++) {
            var types = setsToReplace_1[_i];
            _loop_1(types);
        }
        return _this;
    }
    GraphRewriteBuilder.prototype.registerUnion = function (typeRefs, reconstituted) {
        var set = immutable_1.Set(typeRefs);
        Support_1.assert(!this._reconstitutedUnions.has(set), "Cannot register reconstituted set twice");
        this._reconstitutedUnions = this._reconstitutedUnions.set(set, reconstituted);
    };
    GraphRewriteBuilder.prototype.followIndex = function (index) {
        var entry = this.types[index];
        if (typeof entry === "number") {
            return this.followIndex(entry);
        }
        return index;
    };
    GraphRewriteBuilder.prototype.typeForEntry = function (entry) {
        if (typeof entry === "number") {
            entry = this.types[this.followIndex(entry)];
            if (typeof entry === "number") {
                return Support_1.panic("followIndex led us to a forwarding entry");
            }
        }
        return entry;
    };
    GraphRewriteBuilder.prototype.withForwardingRef = function (maybeForwardingRef, typeCreator) {
        if (maybeForwardingRef !== undefined) {
            return typeCreator(maybeForwardingRef);
        }
        var forwardingRef = new TypeRef(this.typeGraph, undefined, this);
        var actualRef = typeCreator(forwardingRef);
        forwardingRef.resolve(actualRef);
        return actualRef;
    };
    GraphRewriteBuilder.prototype.replaceSet = function (typesToReplace, maybeForwardingRef) {
        var _this = this;
        return this.withForwardingRef(maybeForwardingRef, function (forwardingRef) {
            typesToReplace.forEach(function (t) {
                var originalRef = t.typeRef;
                var index = originalRef.getIndex();
                _this._reconstitutedTypes = _this._reconstitutedTypes.set(index, forwardingRef);
                _this._setsToReplaceByMember = _this._setsToReplaceByMember.remove(index);
            });
            return _this._replacer(typesToReplace, _this, forwardingRef);
        });
    };
    GraphRewriteBuilder.prototype.forceReconstituteTypeRef = function (originalRef, maybeForwardingRef) {
        var _this = this;
        return this.withForwardingRef(maybeForwardingRef, function (forwardingRef) {
            _this._reconstitutedTypes = _this._reconstitutedTypes.set(originalRef.getIndex(), forwardingRef);
            var _a = originalRef.deref(), originalType = _a[0], originalNames = _a[1];
            return originalType.map(new TypeReconstituter(_this, _this.alphabetizeProperties, originalNames, forwardingRef), function (ref) { return _this.reconstituteTypeRef(ref); });
        });
    };
    GraphRewriteBuilder.prototype.assertTypeRefsToReconstitute = function (typeRefs, forwardingRef) {
        for (var _i = 0, typeRefs_1 = typeRefs; _i < typeRefs_1.length; _i++) {
            var originalRef = typeRefs_1[_i];
            Support_1.assert(originalRef.graph === this._originalGraph, "Trying to reconstitute a type from the wrong graph");
        }
        if (forwardingRef !== undefined) {
            Support_1.assert(forwardingRef.graph === this.typeGraph, "Trying to forward a type to the wrong graph");
        }
    };
    GraphRewriteBuilder.prototype.reconstituteTypeRef = function (originalRef, maybeForwardingRef) {
        var maybeRef = this.lookupTypeRefs([originalRef], maybeForwardingRef);
        if (maybeRef !== undefined) {
            return maybeRef;
        }
        return this.forceReconstituteTypeRef(originalRef, maybeForwardingRef);
    };
    GraphRewriteBuilder.prototype.reconstituteType = function (t, forwardingRef) {
        return this.reconstituteTypeRef(t.typeRef, forwardingRef);
    };
    // If the union of these type refs have been, or are supposed to be, reconstituted to
    // one target type, return it.  Otherwise return undefined.
    GraphRewriteBuilder.prototype.lookupTypeRefs = function (typeRefs, forwardingRef) {
        this.assertTypeRefsToReconstitute(typeRefs, forwardingRef);
        // Check whether we have already reconstituted them.  That means ensuring
        // that they all have the same target type.
        var maybeRef = this._reconstitutedTypes.get(typeRefs[0].getIndex());
        if (maybeRef !== undefined && maybeRef !== forwardingRef) {
            var allEqual = true;
            for (var i = 1; i < typeRefs.length; i++) {
                if (this._reconstitutedTypes.get(typeRefs[i].getIndex()) !== maybeRef) {
                    allEqual = false;
                    break;
                }
            }
            if (allEqual) {
                if (forwardingRef !== undefined) {
                    forwardingRef.resolve(maybeRef);
                }
                return maybeRef;
            }
        }
        // Has this been reconstituted as a set?
        maybeRef = this._reconstitutedUnions.get(immutable_1.Set(typeRefs));
        if (maybeRef !== undefined && maybeRef !== forwardingRef) {
            if (forwardingRef !== undefined) {
                forwardingRef.resolve(maybeRef);
            }
            return maybeRef;
        }
        // Is this set requested to be replaced?  If not, we're out of options.
        var maybeSet = this._setsToReplaceByMember.get(typeRefs[0].getIndex());
        if (maybeSet === undefined) {
            return undefined;
        }
        for (var i = 1; i < typeRefs.length; i++) {
            if (this._setsToReplaceByMember.get(typeRefs[i].getIndex()) !== maybeSet) {
                return undefined;
            }
        }
        // Yes, this set is requested to be replaced, so do it.
        return this.replaceSet(maybeSet, forwardingRef);
    };
    GraphRewriteBuilder.prototype.finish = function () {
        var _this = this;
        this._originalGraph.topLevels.forEach(function (t, name) {
            _this.addTopLevel(name, _this.reconstituteType(t));
        });
        return _super.prototype.finish.call(this);
    };
    GraphRewriteBuilder.prototype.setLostTypeAttributes = function () {
        this._lostTypeAttributes = true;
    };
    Object.defineProperty(GraphRewriteBuilder.prototype, "lostTypeAttributes", {
        get: function () {
            return this._lostTypeAttributes;
        },
        enumerable: true,
        configurable: true
    });
    return GraphRewriteBuilder;
}(TypeBuilder));
exports.GraphRewriteBuilder = GraphRewriteBuilder;
function addAttributes(accumulatorAttributes, newAttributes) {
    if (accumulatorAttributes === undefined)
        return newAttributes;
    return TypeAttributes_1.combineTypeAttributes(accumulatorAttributes, newAttributes);
}
function setAttributes(attributeMap, kind, newAttributes) {
    return attributeMap.set(kind, addAttributes(attributeMap.get(kind), newAttributes));
}
function moveAttributes(map, fromKind, toKind) {
    var fromAttributes = Support_1.defined(map.get(fromKind));
    map = map.remove(fromKind);
    return setAttributes(map, toKind, fromAttributes);
}
var UnionAccumulator = /** @class */ (function () {
    function UnionAccumulator(_conflateNumbers) {
        this._conflateNumbers = _conflateNumbers;
        this._nonStringTypeAttributes = immutable_1.OrderedMap();
        this._stringTypeAttributes = immutable_1.OrderedMap();
        this.arrayData = [];
        this.mapData = [];
        this.classData = [];
        this._lostTypeAttributes = false;
        // FIXME: we're losing order here
        this.enumCaseMap = {};
        this.enumCases = [];
    }
    UnionAccumulator.prototype.have = function (kind) {
        return (this._nonStringTypeAttributes.has(kind) || this._stringTypeAttributes.has(kind));
    };
    Object.defineProperty(UnionAccumulator.prototype, "haveString", {
        get: function () {
            return this.have("string");
        },
        enumerable: true,
        configurable: true
    });
    UnionAccumulator.prototype.addAny = function (attributes) {
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "any", attributes);
        this._lostTypeAttributes = true;
    };
    UnionAccumulator.prototype.addNull = function (attributes) {
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "null", attributes);
    };
    UnionAccumulator.prototype.addBool = function (attributes) {
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "bool", attributes);
    };
    UnionAccumulator.prototype.addInteger = function (attributes) {
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "integer", attributes);
    };
    UnionAccumulator.prototype.addDouble = function (attributes) {
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "double", attributes);
    };
    UnionAccumulator.prototype.addStringType = function (kind, attributes) {
        if (this.have(kind)) {
            this._stringTypeAttributes = setAttributes(this._stringTypeAttributes, kind, attributes);
            return;
        }
        // string overrides all other string types, as well as enum
        if (kind === "string") {
            var oldAttributes = TypeAttributes_1.combineTypeAttributes(this._stringTypeAttributes.valueSeq().toArray());
            var newAttributes = addAttributes(oldAttributes, attributes);
            this._stringTypeAttributes = this._stringTypeAttributes.clear().set(kind, newAttributes);
            this.enumCaseMap = {};
            this.enumCases = [];
        }
        else {
            this._stringTypeAttributes = setAttributes(this._stringTypeAttributes, kind, attributes);
        }
    };
    UnionAccumulator.prototype.addArray = function (t, attributes) {
        this.arrayData.push(t);
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "array", attributes);
    };
    UnionAccumulator.prototype.addClass = function (t, attributes) {
        this.classData.push(t);
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "class", attributes);
    };
    UnionAccumulator.prototype.addMap = function (t, attributes) {
        this.mapData.push(t);
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "map", attributes);
    };
    UnionAccumulator.prototype.addEnumCases = function (cases, attributes) {
        var _this = this;
        if (this.have("string")) {
            this.addStringType("string", attributes);
            return;
        }
        cases.forEach(function (count, s) {
            if (!Object.prototype.hasOwnProperty.call(_this.enumCaseMap, s)) {
                _this.enumCaseMap[s] = 0;
                _this.enumCases.push(s);
            }
            _this.enumCaseMap[s] += count;
        });
        this._stringTypeAttributes = setAttributes(this._stringTypeAttributes, "enum", attributes);
    };
    UnionAccumulator.prototype.addEnumCase = function (s, count, attributes) {
        this.addEnumCases(immutable_1.OrderedMap([[s, count]]), attributes);
    };
    UnionAccumulator.prototype.getMemberKinds = function () {
        var merged = this._nonStringTypeAttributes.merge(this._stringTypeAttributes);
        if (merged.isEmpty()) {
            return immutable_1.OrderedMap([["none", immutable_1.Map()]]);
        }
        if (this._nonStringTypeAttributes.has("any")) {
            Support_1.assert(this._lostTypeAttributes, "This had to be set when we added 'any'");
            var allAttributes = TypeAttributes_1.combineTypeAttributes(merged.valueSeq().toArray());
            return immutable_1.OrderedMap([["any", allAttributes]]);
        }
        if (this._conflateNumbers && this.have("integer") && this.have("double")) {
            merged = moveAttributes(merged, "integer", "double");
        }
        if (this.have("map")) {
            merged = moveAttributes(merged, "map", "class");
        }
        return merged;
    };
    Object.defineProperty(UnionAccumulator.prototype, "lostTypeAttributes", {
        get: function () {
            return this._lostTypeAttributes;
        },
        enumerable: true,
        configurable: true
    });
    return UnionAccumulator;
}());
exports.UnionAccumulator = UnionAccumulator;
var FauxUnion = /** @class */ (function () {
    function FauxUnion() {
    }
    FauxUnion.prototype.getAttributes = function () {
        return TypeAttributes_1.emptyTypeAttributes;
    };
    return FauxUnion;
}());
function attributesForTypes(types) {
    var unionsForType = immutable_1.OrderedMap();
    var typesForUnion = immutable_1.Map();
    var unions = immutable_1.OrderedSet();
    var unionsEquivalentToRoot = immutable_1.Set();
    function traverse(t, path, isEquivalentToRoot) {
        if (t instanceof Type_1.UnionType) {
            unions = unions.add(t);
            if (isEquivalentToRoot) {
                unionsEquivalentToRoot = unionsEquivalentToRoot.add(t);
            }
            path = path.add(t);
            isEquivalentToRoot = isEquivalentToRoot && t.members.size === 1;
            t.members.forEach(function (m) { return traverse(m, path, isEquivalentToRoot); });
        }
        else {
            unionsForType = unionsForType.update(t, immutable_1.Set(), function (s) { return s.union(path); });
            path.forEach(function (u) {
                typesForUnion = typesForUnion.update(u, immutable_1.Set(), function (s) { return s.add(t); });
            });
        }
    }
    var rootPath = immutable_1.Set([new FauxUnion()]);
    types.forEach(function (t) { return traverse(t, rootPath, types.size === 1); });
    var resultAttributes = unionsForType.map(function (unionForType, t) {
        var singleAncestors = unionForType.filter(function (u) { return Support_1.defined(typesForUnion.get(u)).size === 1; });
        Support_1.assert(singleAncestors.every(function (u) { return Support_1.defined(typesForUnion.get(u)).has(t); }), "We messed up bookkeeping");
        var inheritedAttributes = singleAncestors.toArray().map(function (u) { return u.getAttributes(); });
        return TypeAttributes_1.combineTypeAttributes([t.getAttributes()].concat(inheritedAttributes));
    });
    var unionAttributes = unions.toArray().map(function (u) {
        var t = typesForUnion.get(u);
        if (t !== undefined && t.size === 1) {
            return TypeAttributes_1.emptyTypeAttributes;
        }
        var attributes = u.getAttributes();
        if (unionsEquivalentToRoot.has(u)) {
            return attributes;
        }
        return TypeAttributes_1.makeTypeAttributesInferred(attributes);
    });
    return [resultAttributes, TypeAttributes_1.combineTypeAttributes(unionAttributes)];
}
// FIXME: Move this to UnifyClasses.ts?
var TypeRefUnionAccumulator = /** @class */ (function (_super) {
    __extends(TypeRefUnionAccumulator, _super);
    function TypeRefUnionAccumulator() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    // There is a method analogous to this in the IntersectionAccumulator.  It might
    // make sense to find a common interface.
    TypeRefUnionAccumulator.prototype.addType = function (t, attributes) {
        var _this = this;
        Type_1.matchTypeExhaustive(t, function (_noneType) {
            return;
        }, function (_anyType) { return _this.addAny(attributes); }, function (_nullType) { return _this.addNull(attributes); }, function (_boolType) { return _this.addBool(attributes); }, function (_integerType) { return _this.addInteger(attributes); }, function (_doubleType) { return _this.addDouble(attributes); }, function (stringType) {
            var enumCases = stringType.enumCases;
            if (enumCases === undefined) {
                _this.addStringType("string", attributes);
            }
            else {
                _this.addEnumCases(enumCases, attributes);
            }
        }, function (arrayType) { return _this.addArray(arrayType.items.typeRef, attributes); }, function (classType) { return _this.addClass(classType.typeRef, attributes); }, function (mapType) { return _this.addMap(mapType.values.typeRef, attributes); }, 
        // FIXME: We're not carrying counts, so this is not correct if we do enum
        // inference.  JSON Schema input uses this case, however, without enum
        // inference, which is fine, but still a bit ugly.
        function (enumType) { return _this.addEnumCases(enumType.cases.toOrderedMap().map(function (_) { return 1; }), attributes); }, function (_unionType) {
            return Support_1.panic("The unions should have been eliminated in attributesForTypesInUnion");
        }, function (_dateType) { return _this.addStringType("date", attributes); }, function (_timeType) { return _this.addStringType("time", attributes); }, function (_dateTimeType) { return _this.addStringType("date-time", attributes); });
    };
    TypeRefUnionAccumulator.prototype.addTypes = function (types) {
        var _this = this;
        var _a = attributesForTypes(types), attributesMap = _a[0], unionAttributes = _a[1];
        attributesMap.forEach(function (attributes, t) { return _this.addType(t, attributes); });
        return unionAttributes;
    };
    return TypeRefUnionAccumulator;
}(UnionAccumulator));
exports.TypeRefUnionAccumulator = TypeRefUnionAccumulator;
var UnionBuilder = /** @class */ (function () {
    function UnionBuilder(typeBuilder) {
        this.typeBuilder = typeBuilder;
    }
    UnionBuilder.prototype.makeTypeOfKind = function (typeProvider, kind, typeAttributes, forwardingRef) {
        switch (kind) {
            case "any":
            case "none":
            case "null":
            case "bool":
            case "double":
            case "integer":
            case "date":
            case "time":
            case "date-time":
                var t = this.typeBuilder.getPrimitiveType(kind, forwardingRef);
                this.typeBuilder.addAttributes(t, typeAttributes);
                return t;
            case "string":
                return this.typeBuilder.getStringType(typeAttributes, undefined, forwardingRef);
            case "enum":
                return this.makeEnum(typeProvider.enumCases, typeProvider.enumCaseMap, typeAttributes, forwardingRef);
            case "class":
                return this.makeClass(typeProvider.classData, typeProvider.mapData, typeAttributes, forwardingRef);
            case "array":
                return this.makeArray(typeProvider.arrayData, typeAttributes, forwardingRef);
            default:
                if (kind === "union" || kind === "map" || kind === "intersection") {
                    return Support_1.panic("getMemberKinds() shouldn't return " + kind);
                }
                return Support_1.assertNever(kind);
        }
    };
    UnionBuilder.prototype.buildUnion = function (typeProvider, unique, typeAttributes, forwardingRef) {
        var _this = this;
        var kinds = typeProvider.getMemberKinds();
        if (typeProvider.lostTypeAttributes) {
            this.typeBuilder.setLostTypeAttributes();
        }
        if (kinds.size === 1) {
            var _a = kinds.toArray()[0], kind = _a[0], memberAttributes = _a[1];
            var allAttributes = TypeAttributes_1.combineTypeAttributes(typeAttributes, memberAttributes);
            var t = this.makeTypeOfKind(typeProvider, kind, allAttributes, forwardingRef);
            return t;
        }
        var union = unique
            ? this.typeBuilder.getUniqueUnionType(typeAttributes, undefined, forwardingRef)
            : undefined;
        var types = [];
        kinds.forEach(function (memberAttributes, kind) {
            types.push(_this.makeTypeOfKind(typeProvider, kind, memberAttributes, undefined));
        });
        var typesSet = immutable_1.OrderedSet(types);
        if (union !== undefined) {
            this.typeBuilder.setSetOperationMembers(union, typesSet);
            return union;
        }
        else {
            return this.typeBuilder.getUnionType(typeAttributes, typesSet, forwardingRef);
        }
    };
    return UnionBuilder;
}());
exports.UnionBuilder = UnionBuilder;
